/**********************************************************************************/
/*                                                                                */
/* File:   draw.cpp                                                               */
/* Author: bkenwright@xbdev.net                                                   */
/* URL:    www.xbdev.net                                                          */
/* Date:   21-12-2005 (xmas)                                                      */
/*                                                                                */
/**********************************************************************************/



#include <windows.h>				// Windows header files
#include <assert.h>					// assert(..)


#pragma comment(lib, "d3dx9.lib")	// directx library files
#pragma comment(lib, "d3d9.lib")

#include <d3d9.h>					// directx header files
#include <d3dx9.h>
#include <d3dx9mesh.h>  

#include <stdio.h>					//sprintf(..)

#include "dxdbg.h"

#include "common.h"					// SafeRelease(..)

#include "bsp.h"


#define	 BUILD_VERSION				"0.0.01a"

#define  SZ_FILE_BSP				"test_level.bsp"		// Quake 3 BSP Level File

		


LPDIRECT3D9							g_pD3D					= NULL;
LPDIRECT3DDEVICE9					g_pd3dDevice			= NULL;

HWND								g_hwnd					= 0;

static float						g_fSpinX				= 0.0f;
static float						g_fSpinY				= 0.0f;
static float						g_fZoom					= -200.0f;

static bool							g_bWire					= false;
static bool							g_bTopView				= true;
static bool							g_bDrawShadowLevel		= true;

static int							g_nTypeCheck			= 1;

static float						g_fLine					= 500.0f;

static int							g_nTotalTris			= 0;
static int							g_nDrawnTris			= 0;

static D3DXVECTOR3					g_vCamPosition			= D3DXVECTOR3(0,0,100);
static D3DXVECTOR3					g_vCamDir				= D3DXVECTOR3(0,1,0);

//--------------------------------------------------------------------------------//
//        Creation/Destruction of our DirectX Applications                        //
//--------------------------------------------------------------------------------//

bool Create(HWND hWnd)
{
	g_hwnd = hWnd;
    //First of all, create the main D3D object. If it is created successfully 
	//we should get a pointer to an IDirect3D8 interface.
    g_pD3D = Direct3DCreate9(D3D_SDK_VERSION);
	if( g_pD3D==NULL )
	{
		ErrMsg("Error Creating DirectX Object Instance");
		return false;
	}

    //Get the current display mode
    D3DDISPLAYMODE d3ddm;
	HRESULT hr = 
    g_pD3D->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &d3ddm);
	if( hr!=S_OK )
	{
		ErrMsg("Failed Getting Display Mode");
		return false;
	}

    //Create a structure to hold the settings for our device
    D3DPRESENT_PARAMETERS d3dpp; 
    ZeroMemory(&d3dpp, sizeof(d3dpp));
    //Fill the structure. 
    //We want our program to be windowed, and set the back buffer to a format
    //that matches our current display mode
    d3dpp.Windowed				 = TRUE;
    d3dpp.SwapEffect			 = D3DSWAPEFFECT_DISCARD;
    d3dpp.BackBufferFormat		 = d3ddm.Format;
    //For depth buffering (e.g.) the z-buffer
    d3dpp.BackBufferCount		 = 1;
    d3dpp.AutoDepthStencilFormat = D3DFMT_D16;
    d3dpp.EnableAutoDepthStencil = TRUE;
    
	//Create a Direct3D device
	hr =														// Error Checking
    g_pD3D->CreateDevice(D3DADAPTER_DEFAULT,					// UINT Adapter
						 D3DDEVTYPE_HAL,						// D3DDEVTYPE DeviceType
						 hWnd,									// HWND hFocusWindow
						 D3DCREATE_SOFTWARE_VERTEXPROCESSING,	// DWORD BehaviorFlags
						 &d3dpp,								// D3DPRESENT_PARAMETERS *
						 &g_pd3dDevice);						// IDirect3DDevice9 *
	if( hr!=S_OK )
	{
		ErrMsg("Failed Creating Rendering Device");
		return false;
	}


    //Turn on z-buffering
    g_pd3dDevice->SetRenderState(D3DRS_ZENABLE, D3DZB_TRUE);

	//Set our culling mode to Counter Clock Wise
	g_pd3dDevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_CCW);



	return true;
}//End Create(..)


bool Release()
{
	SafeRelease( g_pd3dDevice );

	SafeRelease( g_pD3D );

	return true;
}//End Release()

//--------------------------------------------------------------------------------//

// Capture mouse input, and use it to make our demo more interactive
void MouseActions(int x,     // x mouse position
				  int y,     // y mouse position
				  int lmb,   // 0 None, 1 Button Down, 2 Button Released
				  int rmb,   // 0 None, 1 Button Down, 2 Button Released
				  int mmb)   // 0 None, 1 Button Down, 2 Button Released
{
	static int oldy = x;
	static int oldx = y;

	static bool bSpin = false;
	static bool bZoom = false;

	int dx = oldx - x;
	int dy = oldy - y;

	oldx = x;
	oldy = y;

	if( lmb== 1 )
		bSpin = true;
	if( lmb==2 )
		bSpin = false;

	if( rmb==1 )
		bZoom = true;
	if( rmb==2 )
		bZoom = false;
		

	if( bSpin )
	{
		g_fSpinX+=0.5f*dx;
		g_fSpinY+=0.5f*dy;
	}

	if( bZoom )
		g_fZoom+=0.5f*dy;
}



void KeyBoard()
{
	if( GetKeyState('W')		& 0x80 )	g_bWire = !g_bWire;
	if( GetKeyState('V')		& 0x80 )	g_bTopView = !g_bTopView;

	if( GetKeyState('1')		& 0x80 )	g_nTypeCheck = 1;
	if( GetKeyState('2')		& 0x80 )	g_nTypeCheck = 2;
	if( GetKeyState('3')		& 0x80 )	g_nTypeCheck = 3;
	if( GetKeyState('4')		& 0x80 )	g_nTypeCheck = 4;


	if( GetKeyState('S')		& 0x80 )	g_bDrawShadowLevel=!g_bDrawShadowLevel;

	if( (g_nTypeCheck==2) || 
		(g_nTypeCheck==4) || 
		(g_bTopView==false)  )
	{
		// Move using the cameras direction
		if( GetKeyState(VK_UP)		& 0x80 )	g_vCamPosition += g_vCamDir*3.0f;
		if( GetKeyState(VK_DOWN)	& 0x80 )	g_vCamPosition -= g_vCamDir*3.0f;
		if( GetKeyState(VK_RIGHT)	& 0x80 )	g_fSpinX-=2.0f;
		if( GetKeyState(VK_LEFT)	& 0x80 )	g_fSpinX+=2.0f;

	}
	else
	{
		
		// Basic left/right up/down movement
		if( GetKeyState(VK_UP)		& 0x80 )	g_vCamPosition.y -= 3.0f;
		if( GetKeyState(VK_DOWN)	& 0x80 )	g_vCamPosition.y += 3.0f;
		if( GetKeyState(VK_RIGHT)	& 0x80 )	g_vCamPosition.x += 3.0f;
		if( GetKeyState(VK_LEFT)	& 0x80 )	g_vCamPosition.x -= 3.0f;
		
	}

	if( GetKeyState('Q')		& 0x80 )	g_fLine +=5.0f;
	if( GetKeyState('A')		& 0x80 )	g_fLine -=5.0f;
}



//-------------------------------------------------------------------------//

void SetMaterial()
{
      D3DMATERIAL9 matMaterial;
      D3DCOLORVALUE rgbaDiffuse  = {1.0, 1.0, 1.0, 0.0};  //RGBA
      D3DCOLORVALUE rgbaAmbient  = {1.0, 1.0, 1.0, 0.0};
      D3DCOLORVALUE rgbaSpecular = {0.2f, 0.2f, 0.2f, 0.0};
      // emissive is for a glow effect
      D3DCOLORVALUE rgbaEmissive = {0.0, 0.0, 0.0, 0.0};

      matMaterial.Diffuse  = rgbaDiffuse;
      matMaterial.Ambient  = rgbaAmbient;
      matMaterial.Specular = rgbaSpecular;
      matMaterial.Emissive = rgbaEmissive;
      matMaterial.Power    = 500.0f;

      g_pd3dDevice->SetMaterial(&matMaterial);
}//End SetMaterial(..)

 

void SetLight()
{
	//Set a light up?
    D3DLIGHT9 d3dLight;

    //Initialize the light structure.
    ZeroMemory(&d3dLight, sizeof(D3DLIGHT9));

    //Set up a white point light at (0, 0, -10).
    d3dLight.Type			= D3DLIGHT_POINT;

    // Actual colour of the light. (e.g. white).
    d3dLight.Diffuse.r		= 1.0f;
    d3dLight.Diffuse.g		= 1.0f;
    d3dLight.Diffuse.b		= 1.0f;

    // Colour when no light hits the object.
    d3dLight.Ambient.r		= 0.2f;
    d3dLight.Ambient.g		= 0.2f;
    d3dLight.Ambient.b		= 0.2f;

    // Shiny effect.
    d3dLight.Specular.r		= 0.3f;
    d3dLight.Specular.g     = 0.3f;
    d3dLight.Specular.b     = 0.3f;

    d3dLight.Position.x     = 0.0f;
    d3dLight.Position.y     = 0.0f;
    d3dLight.Position.z     = -30.0f;
	//d3dLight.Position	 = g_vCamPosition;

    d3dLight.Attenuation0	= 1.0f; 
    d3dLight.Attenuation1	= 0.0f; 
    d3dLight.Attenuation2	= 0.0f; 
    d3dLight.Range			= 200.0f;

    g_pd3dDevice->SetLight(0, &d3dLight);
    g_pd3dDevice->LightEnable(0,TRUE);

    //Turn on lighting
    g_pd3dDevice->SetRenderState(D3DRS_LIGHTING, TRUE);
    g_pd3dDevice->SetRenderState(D3DRS_AMBIENT, 
								 D3DCOLOR_XRGB(50,50,50));
}//End SetLight(..)



void SetCameraTopView()
{
	// Do camera movement around the level

	// Height off the ground
	static float fHeight = 40.0f;

	D3DXVECTOR3 vRefCamDir = D3DXVECTOR3(0,1,0);

	// Limit the amount the player can look up or down
	if( g_fSpinY>  10 ) g_fSpinY=10;
	if( g_fSpinY< -10 ) g_fSpinY=-10;

	D3DXMATRIX matRot;
	D3DXMatrixRotationYawPitchRoll( &matRot,					// Pointer to the D3DXMATRIX result 
									0,							// Yaw around the y-axis, in radians
									-D3DXToRadian(g_fSpinY),	// Pitch around the x-axis, in radians
									-D3DXToRadian(g_fSpinX));	// Roll around the z-axis, in radians

	D3DXVec3TransformCoord(&g_vCamDir, &vRefCamDir, &matRot);

	//g_vCamPosition.y = fHeight;
	
	D3DXVECTOR3 vCamAt = g_vCamDir + g_vCamPosition;
	//vCamAt.y = fHeight;

    // [1] D3DTS_VIEW
    D3DXMATRIX v;
    g_pd3dDevice->SetTransform(	D3DTS_VIEW, 
								D3DXMatrixLookAtLH(&v, 
								&D3DXVECTOR3(0,0.5f,2400),	// pEye
                                &D3DXVECTOR3(0,0,0),		// pAt
                                &D3DXVECTOR3(0,0,1)));		// pUp

    // [2] D3DTS_PROJECTION
    D3DXMATRIX p;
    g_pd3dDevice->SetTransform( D3DTS_PROJECTION, 
								D3DXMatrixPerspectiveFovLH( &p,  
								D3DX_PI/4,  
								1.0f,
                                1.0f,2500.0f));

	// [3] 	D3DTS_WORLD
	D3DXMATRIX matScale;
	D3DXMatrixScaling(&matScale,
					  1.0f,		// x scale
					  1.0f,		// y scale
					  1.0f);	// z scale
	g_pd3dDevice->SetTransform( D3DTS_WORLD, 
								&(matScale));
}//End SetCamera(..)

void SetCameraLevelView()
{
	// Do camera movement around the level

	// Height off the ground
	static float fHeight = 40.0f;

	D3DXVECTOR3 vRefCamDir = D3DXVECTOR3(0,1,0);

	// Limit the amount the player can look up or down
	if( g_fSpinY>  10 ) g_fSpinY=10;
	if( g_fSpinY< -10 ) g_fSpinY=-10;


	D3DXMATRIX matRot;
	D3DXMatrixRotationYawPitchRoll( &matRot,					// Pointer to the D3DXMATRIX result 
									0,							// Yaw around the y-axis, in radians
									-D3DXToRadian(g_fSpinY),	// Pitch around the x-axis, in radians
									-D3DXToRadian(g_fSpinX));	// Roll around the z-axis, in radians

	D3DXVec3TransformCoord(&g_vCamDir, &vRefCamDir, &matRot);

	g_vCamPosition.z = fHeight;
	
	D3DXVECTOR3 vCamAt = g_vCamDir + g_vCamPosition;
	//vCamAt.y = fHeight;

    // [1] D3DTS_VIEW
    D3DXMATRIX v;
    g_pd3dDevice->SetTransform(	D3DTS_VIEW, 
								D3DXMatrixLookAtLH(&v, 
								&g_vCamPosition,			// pEye
                                &vCamAt,					// pAt
                                &D3DXVECTOR3(0,0,1)));		// pUp

    // [2] D3DTS_PROJECTION
    D3DXMATRIX p;
    g_pd3dDevice->SetTransform( D3DTS_PROJECTION, 
								D3DXMatrixPerspectiveFovLH( &p,  
								D3DX_PI/4,  
								1.0f,
                                1.0f,2500.0f));

	// [3] 	D3DTS_WORLD
	D3DXMATRIX matScale;
	D3DXMatrixScaling(&matScale,
					  1.0f,		// x scale
					  1.0f,		// y scale
					  1.0f);	// z scale
	g_pd3dDevice->SetTransform( D3DTS_WORLD, 
								&(matScale));
}//End SetCamera(..)

////////////////////////////////////////////////////////////////////////////////////
// Display information to the user on the screen
void RenderInfo()
{
	// Using a static variable so we don't need a global
	static debugtxt txt(g_pd3dDevice);

	// Get the window size
	RECT rect;
	GetClientRect(g_hwnd, &rect);

	int yl = rect.bottom - 20;		// y left screen pos text

	char buf[200]; // large temp txt buffer
	sprintf(buf, "X-Rot:%.2f  Y-Rot:%.2f", g_fSpinX, g_fSpinY);
	txt.printf(buf, 10, yl ); yl-=20;

	sprintf(buf, "X-Pos:%.2f  Y-Pos:%.2f  Z-Pos:%.2f", 
				 g_vCamPosition.x, g_vCamPosition.y, g_vCamPosition.z);
	txt.printf(buf, 10, yl ); yl-=20;

	sprintf(buf, "Line: %.2f", g_fLine);
	txt.printf(buf, 10, yl ); yl-=20;

	sprintf(buf, "Total Tris: %d", g_nTotalTris);
	txt.printf(buf, 10, yl ); yl-=20;

	sprintf(buf, "Drawn Tris: %d", g_nDrawnTris);
	txt.printf(buf, 10, yl ); yl-=20;

	int yr = rect.bottom - 20;		// y right screen pos text
	int xr = rect.right  - 150;		// x right screen pos text

	sprintf(buf, "Compiled: %s", __DATE__);
	txt.printf(buf, xr, yr ); yr-=20;

	sprintf(buf, "Build: %s", BUILD_VERSION);
	txt.printf(buf, xr, yr ); yr-=20;

	int ytr = rect.top + 10;		// y top right
	int xtr = rect.right - 300;		// x top right

	sprintf(buf, "'SPACE BAR' Point or Line BSP Checking");
	txt.printf(buf, xtr, ytr ); ytr+=20;

	sprintf(buf, "Press '1' Point View Checking");
	txt.printf(buf, xtr, ytr ); ytr+=20;

	sprintf(buf, "Press '2' Point-Line Checking");
	txt.printf(buf, xtr, ytr ); ytr+=20;

	sprintf(buf, "Press '3' PVS VisData Cluster Checking");
	txt.printf(buf, xtr, ytr ); ytr+=20;

	sprintf(buf, "Press '4' Multiple Line Checking");
	txt.printf(buf, xtr, ytr ); ytr+=20;

	sprintf(buf, "Press 'W' WireFrame");
	txt.printf(buf, xtr, ytr ); ytr+=20;

	sprintf(buf, "Press 'S' Toggle Shadow Level View");
	txt.printf(buf, xtr, ytr ); ytr+=20;

	sprintf(buf, "Cursor Keys For Movement");
	txt.printf(buf, xtr, ytr ); ytr+=20;
}


void RenderPosition()
{
	static debugsphere ball(g_pd3dDevice);

	if( g_bTopView )
	ball.Draw( g_pd3dDevice,
			   g_vCamPosition.x, g_vCamPosition.y, g_vCamPosition.z, 
			   10.0f);
}


////////////////////////////////////////////////////////////////////////////////////
//  
//  Render(..)
//
////////////////////////////////////////////////////////////////////////////////////
//
//  Our main render loop, which gets called over and over agian to do our
//  drawing...clears the screen, draws the data, then presents it..        
//
////////////////////////////////////////////////////////////////////////////////////
void Render()
{

	if( g_bTopView )
		SetCameraTopView();
	else
		SetCameraLevelView();

	SetLight();

	SetMaterial();

	KeyBoard();

	if( g_bWire )
		g_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_WIREFRAME );
	else
		g_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_SOLID );


	g_pd3dDevice->Clear( 0, 
						 NULL, 
						 D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, 
						 0x88888888, 
						 1.0f, 
						 0 );
    g_pd3dDevice->BeginScene();
	

	


	static CBSP bsp(g_pd3dDevice, SZ_FILE_BSP);

	g_pd3dDevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_CW);

	if( g_bDrawShadowLevel )				// If we want to draw the whole level in shadow
		if( g_bTopView )					// If where First Person View or Top Down
			bsp.Render(g_pd3dDevice);

	D3DXVECTOR3 vDirNorm    = g_vCamDir;
	D3DXVec3Normalize(&vDirNorm, &vDirNorm);
	D3DXVECTOR3 pNose = g_vCamPosition + (vDirNorm*g_fLine);


	if( g_nTypeCheck==1 )
	{
		bsp.CheckNodeForPoint(g_pd3dDevice,
							 g_vCamPosition);
	}
	else if( g_nTypeCheck==2 )
	{
		bsp.CheckNodeForLine(g_pd3dDevice,
							g_vCamPosition,
							pNose );
	}
	else if( g_nTypeCheck==3 )
	{
		bsp.CheckClusterPoint(g_pd3dDevice, 
							  g_vCamPosition);
	}
	else if( g_nTypeCheck==4 )
	{
		bsp.CheckMultipleLines(	g_pd3dDevice,
					g_vCamPosition,
					g_vCamDir,
					g_fLine);
	}

	g_nTotalTris = bsp.GetTotalTris();
	g_nDrawnTris = bsp.GetDrawnTris();

	RenderInfo();

	RenderPosition();

	TriDisplay(g_pd3dDevice, g_nTotalTris );
	FPSDisplay(g_pd3dDevice);
    g_pd3dDevice->EndScene();
	g_pd3dDevice->Present( NULL, NULL, NULL, NULL );

}//End Render()


